@@ -884,7 +884,7 @@ impl<'a> sop::ReadyWithResult<(Option<sop::SessionKey>, Vec<sop::Verification>)>
             } else {
                 Box::new(io::sink())
             },
-            0,
+            if self.decrypt.verify.certs.is_empty() { 0 } else { 1 },
             self.decrypt.verify.not_before,
             self.decrypt.verify.not_after,
             self.decrypt.verify.certs);
@@ -948,12 +948,30 @@ impl<'a> sop::Ready for ArmorReady<'a> {
 
         let mut source = Generic::new(self.data, None).as_boxed();
 
+        // Peek at the data to see if it is already armored.
+        let armored =
+            source.data(1)?.get(0).map(|b| b & 0x80 == 0).unwrap_or(false);
+
+        // If so, dearmor, and extract the headers.
+        let mut headers: Vec<(String, String)> = vec![];
+        if armored {
+            let mut r = armor::Reader::from_reader(
+                source, armor::ReaderMode::Tolerant(None));
+            headers = r.headers()?.iter().cloned().collect();
+            source = Box::new(Adapter::new(r));
+        }
+
         fn make<'a>(sink: &'a mut (dyn io::Write + Send + Sync),
-                    kind: armor::Kind)
+                    kind: armor::Kind,
+                    headers: Vec<(String, String)>)
                     -> Result<Message>
         {
             let m = Message::new(sink);
-            Armorer::new(m).kind(kind).build().map_err(sop_error)
+            let mut a = Armorer::new(m).kind(kind);
+            for (k, v) in headers {
+                a = a.add_header(k, v);
+            }
+            a.build().map_err(sop_error)
         }
 
         // We make no effort to verify the packet structure.
@@ -982,12 +1000,12 @@ impl<'a> sop::Ready for ArmorReady<'a> {
                 drop(ppr);
                 source = source.into_inner()
                     .expect("the Dup to be popped off");
-                make(sink, kind)?
+                make(sink, kind, headers)?
             },
-            ArmorLabel::Sig => make(sink, armor::Kind::Signature)?,
-            ArmorLabel::Key => make(sink, armor::Kind::SecretKey)?,
-            ArmorLabel::Cert => make(sink, armor::Kind::PublicKey)?,
-            ArmorLabel::Message => make(sink, armor::Kind::Message)?,
+            ArmorLabel::Sig => make(sink, armor::Kind::Signature, headers)?,
+            ArmorLabel::Key => make(sink, armor::Kind::SecretKey, headers)?,
+            ArmorLabel::Cert => make(sink, armor::Kind::PublicKey, headers)?,
+            ArmorLabel::Message => make(sink, armor::Kind::Message, headers)?,
         };
 
         std::io::copy(&mut source, &mut sink)?;
@@ -1028,9 +1046,21 @@ impl sop::Ready for DearmorReady<'_> {
                 -> Result<()> {
         let _ = self.dearmor.sqop;
 
-        // We make no effort to verify the packet structure.
-        let mut r = openpgp::armor::Reader::from_reader(self.data, None);
-        std::io::copy(&mut r, sink)?;
+        // Peek at the data to see if it is already armored.  We make
+        // no effort to verify the packet structure.
+        let mut source = Generic::new(self.data, None).as_boxed();
+        let armored =
+            source.data(1)?.get(0).map(|b| b & 0x80 == 0).unwrap_or(false);
+
+        if armored {
+            // XXX: Give the buffered reader to the armor::Reader once
+            // that is possible.
+            let mut r = openpgp::armor::Reader::from_reader(source, None);
+            std::io::copy(&mut r, sink)?;
+        } else {
+            // Use BufferedReader::copy once available.
+            std::io::copy(&mut source, sink)?;
+        }
         Ok(())
     }
 }
@@ -1715,6 +1745,11 @@ fn print_error_chain(sink: &mut dyn io::Write, err: &anyhow::Error)
 /// XXX: This is a bit of a friction point.  We likely need to improve
 /// the SOP spec a little.
 fn sop_error(e: anyhow::Error) -> sop::Error {
+    let e = match e.downcast::<sop::Error>() {
+        Ok(e) => return e,
+        Err(e) => e,
+    };
+
     let e = match e.downcast::<io::Error>() {
         Ok(e) => return e.into(),
         Err(e) => e,
@@ -1722,7 +1757,7 @@ fn sop_error(e: anyhow::Error) -> sop::Error {
 
     if let Some(e) = e.downcast_ref::<openpgp::Error>() {
         use openpgp::Error::*;
-        return match dbg!(e) {
+        return match e {
             InvalidArgument(_) => sop::Error::BadData,
             InvalidOperation(_) => sop::Error::BadData,
             MalformedPacket(_) => sop::Error::BadData,
@@ -1745,7 +1780,7 @@ fn sop_error(e: anyhow::Error) -> sop::Error {
             BadSignature(_) => sop::Error::BadData,
             MalformedMessage(_) => sop::Error::BadData,
             MalformedCert(_) => sop::Error::BadData,
-            UnsupportedCert(_) => sop::Error::BadData,
+            UnsupportedCert2(_, _) => sop::Error::BadData,
             Expired(_) => sop::Error::BadData,
             NotYetLive(_) => sop::Error::BadData,
             NoBindingSignature(_) => sop::Error::BadData,
